(***********************************************************************)
(*                                                                     *)
(*                           objective Caml                            *)
(*                                                                     *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* $Id: types.mli,v 1.10 1998/11/30 13:06:53 vouillon Exp $ *)

(* Representation of types and declarations *)

open Asttypes

(* Type expressions for the core language *)

(* GENERIC: Tdvar 
     We cannot use the "in-place" generalization, because it changes
     levels of the type variables inside let (or generic) bindings
     by Btype.generic_level. The levels of dynamic type variables
     are important for the generic compilation. Therefore we use
     "copy" generalization function Dtype.generalize_copy, where
     the generalization takes place in the copied types.

     (An idea still not implemented: 
       What we are interesting is just the levels of the dynamic
       type variable. Dtype.generalize_copy may be optimized
       by restricting the coping just the dynamic type variables.)
     ===> IT MUST BE IMPLEMENTED! (AND IMPLEMENTED)

     In the compilation, for each generalized dynamic type variable,
     we have to know its origin, even though it is no longer same as
     the original variable after copied. Therefore each dynamic type
     variable carries a tdvar_info reference which is shared between
     the generalized and original variables.

     tdvar_info tells the status of dynamic variables.
     - Not_generalized 
        The dynamic type variable is never generalized 
     - Generalized
        The dynamic type variable is copy-generalized somewhere.
*)

type tdvar_info =
  | Not_generalized 
  | Generalized

type dynamization_info =
  | Not_exportable
  | Exportable

type type_expr =
  { mutable desc: type_desc; 
    mutable level: int;
    mutable id: int }

and type_desc =
    Tvar
  | Tdvar of tdvar_info ref * dynamization_info ref
  | Tarrow of type_expr * type_expr
  | Ttuple of type_expr list
  | Tconstr of Path.t * type_expr list * abbrev_memo ref
  | Tobject of type_expr * (Path.t * type_expr list) option ref
  | Tfield of string * field_kind * type_expr * type_expr
  | Tnil
  | Tlink of type_expr
  | Tany (* used for generic type patterns *)

and abbrev_memo =
    Mnil
  | Mcons of Path.t * type_expr * type_expr * abbrev_memo
  | Mlink of abbrev_memo ref

and field_kind =
    Fvar of field_kind option ref
  | Fpresent
  | Fabsent

(* Maps of methods and instance variables *)

module Meths : Map.S with type key = string
module Vars  : Map.S with type key = string

(* Value descriptions *)

type value_description =
  { mutable val_type: type_expr;                (* Type of the value *)
    val_kind: value_kind;
    mutable val_konstraints: generic_konstraints
  }

and derived_konstraint = {
    derived_path: Path.t;
    derived_desc: value_description option; (* if recursive, None *)
    derived_type: type_expr;
  } 

and case_konstraint = 
    { konst_case: type_expr;
      konst_derived: derived_konstraint list
    }

and generic_konstraints = (* my name, and constraint *)
  | Konst_Cases of Path.t * case_konstraint list
  | Konst_Derived of Path.t * derived_konstraint list
  | Konst_None

and value_kind =
    Val_reg                             (* Regular value *)
  | Val_prim of Primitive.description   (* Primitive *)
  | Val_ivar of mutable_flag * string   (* Instance variable (mutable ?) *)
  | Val_self of (Ident.t * type_expr) Meths.t ref *
                (Ident.t * Asttypes.mutable_flag * type_expr) Vars.t ref *
                string
                                        (* Self *)
  | Val_anc of (string * Ident.t) list * string
                                        (* Ancestor *)
  | Val_unbound                         (* Unbound variable *)
  | Val_extensible                      (* Extensible generic value *)

(* Constructor descriptions *)

type constructor_description =
  { cstr_res: type_expr;                (* Type of the result *)
    cstr_args: type_expr list;          (* Type of the arguments *)
    cstr_arity: int;                    (* Number of arguments *)
    cstr_tag: constructor_tag;          (* Tag for heap blocks *)
    cstr_consts: int;                   (* Number of constant constructors *)
    cstr_nonconsts: int }               (* Number of non-const constructors *)

and constructor_tag =
    Cstr_constant of int                (* Constant constructor (an int) *)
  | Cstr_block of int                   (* Regular constructor (a block) *)
  | Cstr_exception of Path.t            (* Exception constructor *)

(* Record label descriptions *)

type label_description =
  { lbl_res: type_expr;                 (* Type of the result *)
    lbl_arg: type_expr;                 (* Type of the argument *)
    lbl_mut: mutable_flag;              (* Is this a mutable field? *)
    lbl_pos: int;                       (* Position in block *)
    lbl_all: label_description array;   (* All the labels in this type *)
    lbl_repres: record_representation } (* Representation for this record *)

and record_representation =
    Record_regular                      (* All fields are boxed / tagged *)
  | Record_float                        (* All fields are floats *)

(* Digest stuff *)

type type_digest =
    DLazy
  | DDigest of string
  | DManifest of string Rdigest.rtype
  | DSource of type_digest_source
  | DNone
and path_or_digest =
    Path of Path.t
  | Digest of string
and type_digest_source = {
    regular: Path.t list;
    recursive: Ident.t list;
    declarations: path_or_digest Rdigest.source_declaration list
  }  

(* Type definitions *)

type type_declaration =
  { type_params: type_expr list;
    type_arity: int;
    type_kind: type_kind;
    type_manifest: type_expr option;
    mutable type_digest: type_digest }

and type_kind =
    Type_abstract
  | Type_variant of (string * type_expr list) list
  | Type_record of (string * mutable_flag * type_expr) list

type exception_declaration = type_expr list

(* Type expressions for the class language *)

module Concr : Set.S with type elt = string

type class_type =
    Tcty_constr of Path.t * type_expr list * class_type
  | Tcty_signature of class_signature
  | Tcty_fun of type_expr * class_type

and class_signature =
  { cty_self: type_expr;
    cty_vars: (Asttypes.mutable_flag * type_expr) Vars.t;
    cty_concr: Concr.t }

type class_declaration =
  { cty_params: type_expr list;
    mutable cty_type: class_type;
    cty_path: Path.t;
    cty_new: type_expr option }

type cltype_declaration =
  { clty_params: type_expr list;
    clty_type: class_type;
    clty_path: Path.t }

(* Type expressions for the module language *)

type module_type =
    Tmty_ident of Path.t
  | Tmty_signature of signature
  | Tmty_functor of Ident.t * module_type * module_type

and signature = signature_item list

and signature_item =
    Tsig_value of Ident.t * value_description
  | Tsig_type of Ident.t * type_declaration
  | Tsig_exception of Ident.t * exception_declaration
  | Tsig_module of Ident.t * module_type
  | Tsig_modtype of Ident.t * modtype_declaration
  | Tsig_class of Ident.t * class_declaration
  | Tsig_cltype of Ident.t * cltype_declaration

and modtype_declaration =
    Tmodtype_abstract
  | Tmodtype_manifest of module_type
